
#------------------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Script to install Forefront Identity Manager 2010 R2 Prerequisite Scripts for Reporting
#------------------------------------------------------------------------------------------------------
 
#------------------------------------------------------------------------------------------------------
# Script Parameter Declarations
param
(
    # Data Warehouse machine name
    [parameter(Mandatory=$true)]
    [String]$DataWarehouseServerInstance = "localhost",
            
    # Data Warehouse Repository database Name
    [parameter(Mandatory=$false)]
    [String]$RepositoryDatabaseName = "DWRepository",

    # Data Warehouse StagingAndConfig database Name
    [parameter(Mandatory=$false)]
    [String]$StagingAndConfigDatabaseName = "DWStagingAndConfig",

    # Data Warehouse User name for permissions
    [parameter(Mandatory=$true)]
    [String]$FIMServiceAccountName,

     # Parameter that could be used to override User warning
    [parameter(Mandatory=$false)]
    [bool]$OverrideUserWarning = $false
    
)
# End Script parameter declarations
#------------------------------------------------------------------------------------------------------

#------------------------------------------------------------------------------------------------------
# Script Constant Declarations

# options for Invoke-Sqlcmd commandlet.
Set-Variable -Name DatabaseOption -Option Constant -Value "-Database "
Set-Variable -Name InputFileOption -Option Constant -Value " -InputFile "
Set-Variable -Name OutputSQLErrorsOption -Option Constant -Value " -OutputSQLErrors "
Set-Variable -Name UserNameOption -Option Constant -Value " -Username "
Set-Variable -Name Password -Option Constant -Value "-Password "

# other file name related constants.
Set-Variable -Name PermissionsForMaster -Option Constant -Value "StagingAndConfig\PermissionsForMaster.sql"
Set-Variable -Name PermissionsForStagingAndConfig -Option Constant -Value "StagingAndConfig\PermissionsForStagingAndConfig.sql"
Set-Variable -Name RepositoryPathPrefix -Option Constant -Value "Repository\"
Set-Variable -Name SQLServerCmdletsSnapIn -Option Constant -Value "SqlServerCmdletSnapin100"
Set-Variable -Name SQLServerProviderSnapIn -Option Constant -Value "SqlServerProviderSnapin100"
Set-Variable -Name StagingPathPrefix -Option Constant -Value "StagingAndConfig\"
# End Script Constant Declarations
#------------------------------------------------------------------------------------------------------


#------------------------------------------------------------------------------------------------------
# Function declarations

function Deploy-DbLogin-File()
{        
    try
    {      
        $currentDirectory = Get-Location
    
        $permissionFilePath = Join-Path $currentDirectory $PermissionsForMaster    
        $userNameValue = "UserName=" + $FIMServiceAccountName 
        $databaseNameValue = "StagingAndConfigDatabaseName=" + $StagingAndConfigDatabaseName
        $variableArray = $userNameValue, $databaseNameValue
        Invoke-SqlCmd -ServerInstance $DataWarehouseServerInstance -Variable $variableArray -InputFile $permissionFilePath -AbortOnError
    }
    catch
    {
        $errorString = "Error running Sql file :" + $permissionFilePath
        Write-Host($errorString)
        Write-Host($error)
        throw "Script deployment failed"
    }
}

function Deploy-Permissions-File()
{        
    try
    {      
        $currentDirectory = Get-Location
    
        $permissionFilePath = Join-Path $currentDirectory $PermissionsForMaster    
        $userNameValue = "UserName=" + $FIMServiceAccountName 
        $databaseNameValue = "StagingAndConfigDatabaseName=" + $StagingAndConfigDatabaseName
        $variableArray = $userNameValue, $databaseNameValue
        $permissionFilePath = Join-Path $currentDirectory $PermissionsForStagingAndConfig    

        Invoke-SqlCmd -ServerInstance $DataWarehouseServerInstance -Variable $variableArray -InputFile $permissionFilePath -Database "$StagingAndConfigDatabaseName" -AbortOnError
    }
    catch
    {
        $errorString = "Error running Sql file :" + $permissionFilePath
        Write-Host($errorString)
        Write-Host($error)
        throw "Script deployment failed"
    }

}

function Deploy-Repository-Scripts()
{
    $scriptsForRepositoryDatabase = Scripts-For-Repository
    Deploy-Scripts $RepositoryDatabaseName $scriptsForRepositoryDatabase
}

function Deploy-Staging-And-Config-Scripts()
{
    $scriptsForStagingAndConfigDatabase = Scripts-For-Staging-And-Config
    Deploy-Scripts $StagingAndConfigDatabaseName $scriptsForStagingandConfigDatabase
}


function Deploy-Scripts([String]$databaseToDeployTo, [System.Collections.ArrayList]$scriptFiles)
{      
    foreach($scriptFile in $scriptFiles)
    {
        try
        {      
            $currentDirectory = Get-Location
            $scriptFilePath = Join-Path $currentDirectory $scriptFile	    
            Invoke-SqlCmd -ServerInstance $DataWarehouseServerInstance -Database $databaseToDeployTo -InputFile "$scriptFilePath"  -AbortOnError
        }
        catch
        {
            $errorString = "Error running Sql file :" + $scriptFile
            Write-Host($errorString)
            Write-Host($error)
            throw "Script deployment failed"
        }
    }    
}

function Install-Sql-Server-Snapins
{       
    add-pssnapin $SQLServerCmdletsSnapIn -ErrorAction SilentlyContinue | Out-Null
    add-pssnapin $SQLServerProviderSnapIn -ErrorAction SilentlyContinue | Out-Null
}

function Scripts-For-Repository
{
    [System.Collections.ArrayList]$scriptsForRepository =  @()
           
    $scriptsForRepository.Add($RepositoryPathPrefix + "TransformEntityDim.sql") | Out-Null 
    $scriptsForRepository.Add($RepositoryPathPrefix + "TransformEntityManagedTypeFact.sql") | Out-Null
    $scriptsForRepository.Add($RepositoryPathPrefix + "TransformEntityRelatesToEntityFact.sql") | Out-Null
    $scriptsForRepository.Add($RepositoryPathPrefix + "TransformFIMApprovalDimProc.sql") | Out-Null
    $scriptsForRepository.Add($RepositoryPathPrefix + "TransformFIMApprovalResponseDimProc.sql") | Out-Null
    $scriptsForRepository.Add($RepositoryPathPrefix + "TransformFIMRequestDimProc.sql") | Out-Null
    
    return $scriptsForRepository
}

function Scripts-For-Staging-And-Config
{
   [System.Collections.ArrayList]$scriptsForStagingAndConfig = @()
   
   $scriptsForStagingAndConfig.Add($StagingPathPrefix + "GetMetadata.sql") | Out-Null
   $scriptsForStagingAndConfig.Add($StagingPathPrefix + "StagingTables.sql") | Out-Null
   $scriptsForStagingAndConfig.Add($StagingPathPrefix + "StagingData.sql") | Out-Null
   $scriptsForStagingAndConfig.Add($StagingPathPrefix + "StagingProcedures.sql") | Out-Null
    
   return $scriptsForStagingAndConfig
}
# End Script function declarations
#------------------------------------------------------------------------------------------------------


#------------------------------------------------------------------------------------------------------
# Main Script Logic

try
{
    if($OverrideUserWarning -ne $true)
    {
        Write-Host "It is strongly recommended that you take a backup of all four System Center Databases (ServiceManager, DWStagingAndConfig, DWRepository, DWDataMart)."
        $consoleInputString = Read-Host -prompt "Do you want to continue? (Y/N)"
    }

    if($consoleInputString -eq $null)	
    { 
        $consoleInputString = "Y"
    }

    if($consoleInputString.Equals("Y") -or $consoleInputString.Equals("y") -or $OverrideUserWarning -eq $true)
    {    
        Write-Host "Loading SQL Server Snapins" -foregroundcolor "green"
        Install-Sql-Server-Snapins

        Write-Host "Adding FIMService Login and User" -foregroundcolor "green"
        Deploy-DbLogin-File 

        Write-Host "Deploying DWStagingAndConfiguration Database Scripts" -foregroundcolor "green"
        Deploy-Staging-And-Config-Scripts

        Write-Host "Deploying DWReposity Database Scripts" -foregroundcolor "green"
        Deploy-Repository-Scripts

        Write-Host "Granting Permission to FIMService Login" -foregroundcolor "green"
        Deploy-Permissions-File

        Write-Host "FIM Datawarehouse script installation complete" -backgroundcolor "green" -foregroundcolor "black"
    }
    else
    {
       Write-Host "Script execution stopped. No Scripts have been applied."
    }
}
catch
{
    $errorString = "FIM Datawarehouse script installation FAILED" 
    Write-Host($error)
    Write-Host($errorString) -backgroundcolor "red"
}

# End Main Script Logic
#------------------------------------------------------------------------------------------------------